from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.champion import ChampionData, ChampionListData
from ..dto.champion import ChampionDto, ChampionListDto

T = TypeVar("T")
F = TypeVar("F")


class ChampionTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(ChampionDto, ChampionData)
    def champion_dto_to_data(self, value: ChampionDto, context: PipelineContext = None) -> ChampionData:
        data = deepcopy(value)
        return ChampionData.from_dto(data)

    @transform.register(ChampionListDto, ChampionListData)
    def champion_list_dto_to_data(self, value: ChampionListDto, context: PipelineContext = None) -> ChampionListData:
        data = deepcopy(value)
        data["champions"] = [self.champion_dto_to_data(c) for c in data["champions"]]
        for c in data["champions"]:
            c._update({"region": data["region"]})
        data = data["champions"]
        return ChampionListData(data)



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.match import MatchData, MatchListData, MatchReferenceData, TimelineData, Match, MatchHistory, Timeline
from ..dto.match import MatchDto, MatchListDto, MatchReferenceDto, TimelineDto

T = TypeVar("T")
F = TypeVar("F")


class MatchTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # TODO Integrate `forAccountId` into these transformers.
    # TODO Add all the optional fields to matchlist transformers.

    # Dto to Data

    @transform.register(MatchDto, MatchData)
    def match_dto_to_data(self, value: MatchDto, context: PipelineContext = None) -> MatchData:
        data = deepcopy(value)
        return MatchData.from_dto(data)

    @transform.register(MatchReferenceDto, MatchReferenceData)
    def match_reference_dto_to_data(self, value: MatchReferenceDto, context: PipelineContext = None) -> MatchReferenceData:
        data = deepcopy(value)
        return MatchReferenceData.from_dto(data)

    @transform.register(MatchListDto, MatchListData)
    def matchlist_dto_to_data(self, value: MatchListDto, context: PipelineContext = None) -> MatchListData:
        data = deepcopy(value)
        data = data["matches"]
        return MatchListData([self.match_reference_dto_to_data(match) for match in data], region=value["region"])

    @transform.register(TimelineDto, TimelineData)
    def timeline_dto_to_data(self, value: TimelineDto, context: PipelineContext = None) -> TimelineData:
        data = deepcopy(value)
        return TimelineData.from_dto(data)

    # Data to Core

    @transform.register(MatchData, Match)
    def match_data_to_core(self, value: MatchData, context: PipelineContext = None) -> Match:
        return Match.from_data(value)

    @transform.register(MatchReferenceData, Match)
    def match_reference_data_to_core(self, value: MatchReferenceData, context: PipelineContext = None) -> Match:
        return Match.from_match_reference(value)

    @transform.register(MatchListData, MatchHistory)
    def matchlist_data_to_core(self, value: MatchListData, context: PipelineContext = None) -> MatchHistory:
        return MatchHistory([Match.from_match_reference(ref) for ref in value], region=value.region)

    @transform.register(TimelineData, Timeline)
    def timeline_data_to_core(self, value: TimelineData, context: PipelineContext = None) -> Timeline:
        return Timeline.from_data(value)

    # Core to Dto

    @transform.register(Match, MatchDto)
    def match_core_to_dto(self, value: Match, context: PipelineContext = None) -> MatchDto:
        return value._data[MatchData]._dto

    @transform.register(MatchHistory, MatchListDto)
    def matchlist_core_to_dto(self, value: MatchHistory, context: PipelineContext = None) -> MatchListDto:
        return MatchListDto([self.match_core_to_dto(ref) for ref in value])  # It is very unlikely that this will work. Not going to debug it right now and just hope it isn't necessary / used.

    @transform.register(Timeline, TimelineDto)
    def timeline_core_to_dto(self, value: Timeline, context: PipelineContext = None) -> TimelineDto:
        return value._data[TimelineData]._dto



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.masterypage import MasteryPageData, MasteryPagesData, MasteryPage, MasteryPages
from ..dto.masterypage import MasteryPageDto, MasteryPagesDto

T = TypeVar("T")
F = TypeVar("F")


class MasteriesTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(MasteryPageDto, MasteryPageData)
    def mastery_page_dto_to_data(self, value: MasteryPageDto, context: PipelineContext = None) -> MasteryPageData:
        data = deepcopy(value)
        return MasteryPageData.from_dto(data)

    @transform.register(MasteryPagesDto, MasteryPagesData)
    def mastery_pages_dto_to_data(self, value: MasteryPagesDto, context: PipelineContext = None) -> MasteryPagesData:
        data = deepcopy(value)
        for page in data["pages"]:
            page["region"] = data["region"]
            page["summonerId"] = data["summonerId"]
        data = [self.mastery_page_dto_to_data(page) for page in data["pages"]]
        return MasteryPagesData(data, summoner_id=value["summonerId"])

    # Data to Core

    @transform.register(MasteryPageData, MasteryPage)
    def mastery_page_data_to_core(self, value: MasteryPageData, context: PipelineContext = None) -> MasteryPage:
        data = deepcopy(value)
        return MasteryPage.from_data(data)

    @transform.register(MasteryPagesData, MasteryPages)
    def mastery_pages_data_to_core(self, value: MasteryPagesData, context: PipelineContext = None) -> MasteryPages:
        return MasteryPages([self.mastery_page_data_to_core(page) for page in value], summoner=value.summoner_id)

    # Core to Dto

    @transform.register(MasteryPage, MasteryPageDto)
    def mastery_page_core_to_dto(self, value: MasteryPage, context: PipelineContext = None) -> MasteryPageDto:
        return value._data[MasteryPageData]._dto

    @transform.register(MasteryPages, MasteryPagesDto)
    def mastery_pages_core_to_dto(self, value: MasteryPages, context: PipelineContext = None) -> MasteryPagesDto:
        return MasteryPagesDto({"pages": set([self.mastery_page_core_to_dto(page) for page in value]), "summonerId": value._MasteryPages__summoner.id})



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.summoner import SummonerData, Summoner
from ..dto.summoner import SummonerDto

T = TypeVar("T")
F = TypeVar("F")


class SummonerTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(SummonerDto, SummonerData)
    def summoner_dto_to_data(self, value: SummonerDto, context: PipelineContext = None) -> SummonerData:
        data = deepcopy(value)
        data = SummonerData.from_dto(data)
        return data

    # Data to Core

    @transform.register(SummonerData, Summoner)
    def summoner_data_to_core(self, value: SummonerData, context: PipelineContext = None) -> Summoner:
        return Summoner.from_data(value)

    # Core to Dto

    @transform.register(Summoner, SummonerDto)
    def champion_mastery_core_to_dto(self, value: Summoner, context: PipelineContext = None) -> SummonerDto:
        return value._data[SummonerData]._dto



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.championmastery import ChampionMasteryData, ChampionMasteryListData, ChampionMastery, ChampionMasteries
from ..dto.championmastery import ChampionMasteryDto, ChampionMasteryListDto

T = TypeVar("T")
F = TypeVar("F")


class ChampionMasteryTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(ChampionMasteryDto, ChampionMasteryData)
    def champion_mastery_dto_to_data(self, value: ChampionMasteryDto, context: PipelineContext = None) -> ChampionMasteryData:
        data = deepcopy(value)
        return ChampionMasteryData.from_dto(data)

    @transform.register(ChampionMasteryListDto, ChampionMasteryListData)
    def champion_mastery_list_dto_to_data(self, value: ChampionMasteryListDto, context: PipelineContext = None) -> ChampionMasteryListData:
        data = deepcopy(value)
        data["masteries"] = [self.champion_mastery_dto_to_data(c) for c in data["masteries"]]
        for c in data["masteries"]:
            c._update({"region": data["region"]})
        data = data["masteries"]
        return ChampionMasteryListData(data, region=value["region"], summoner_id=value["summonerId"])

    # Data to Core

    @transform.register(ChampionMasteryData, ChampionMastery)
    def champion_mastery_data_to_core(self, value: ChampionMasteryData, context: PipelineContext = None) -> ChampionMastery:
        return ChampionMastery.from_data(value)

    @transform.register(ChampionMasteryListData, ChampionMasteries)
    def champion_mastery_list_data_to_core(self, value: ChampionMasteryListData, context: PipelineContext = None) -> ChampionMasteries:
        return ChampionMasteries([self.champion_mastery_data_to_core(cm) for cm in value], region=value.region, summoner=value.summoner_id)

    # Core to Dto

    @transform.register(ChampionMastery, ChampionMasteryDto)
    def champion_mastery_core_to_dto(self, value: ChampionMastery, context: PipelineContext = None) -> ChampionMasteryDto:
        return value._data[ChampionMasteryData]._dto

    @transform.register(ChampionMasteries, ChampionMasteryListDto)
    def champion_mastery_list_core_to_dto(self, value: ChampionMasteries, context: PipelineContext = None) -> ChampionMasteryListDto:
        return ChampionMasteryListDto({"masteries": [self.champion_mastery_core_to_dto(cm) for cm in value], "summonerId": value._ChampionMasteries__summoner.id, "region": value.region})



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.status import ShardStatusData, ShardStatus
from ..dto.status import ShardStatusDto

T = TypeVar("T")
F = TypeVar("F")


class StatusTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(ShardStatusDto, ShardStatusData)
    def shard_status_dto_to_data(self, value: ShardStatusDto, context: PipelineContext = None) -> ShardStatusData:
        data = deepcopy(value)
        return ShardStatusData.from_dto(data)

    # Data to Core

    @transform.register(ShardStatusData, ShardStatus)
    def shard_status_data_to_core(self, value: ShardStatusData, context: PipelineContext = None) -> ShardStatus:
        return ShardStatus.from_data(value)

    # Core to Dto

    @transform.register(ShardStatus, ShardStatusDto)
    def shard_status_core_to_dto(self, value: ShardStatus, context: PipelineContext = None) -> ShardStatusDto:
        return value[ShardStatusData]._dto



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.runepage import RunePageData, RunePagesData, RunePage, RunePages

from ..dto.runepage import RunePageDto, RunePagesDto

T = TypeVar("T")
F = TypeVar("F")


class RunesTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(RunePageDto, RunePageData)
    def rune_page_dto_to_data(self, value: RunePageDto, context: PipelineContext = None) -> RunePageData:
        data = deepcopy(value)
        return RunePageData.from_dto(data)

    @transform.register(RunePagesDto, RunePagesData)
    def rune_pages_dto_to_data(self, value: RunePagesDto, context: PipelineContext = None) -> RunePagesData:
        data = deepcopy(value)
        for page in data["pages"]:
            page["region"] = data["region"]
            page["summonerId"] = data["summonerId"]
        data = [self.rune_page_dto_to_data(page) for page in data["pages"]]
        return RunePagesData(data, summoner_id=value["summonerId"])

    # Data to Core

    @transform.register(RunePageData, RunePage)
    def rune_page_data_to_core(self, value: RunePageData, context: PipelineContext = None) -> RunePage:
        data = deepcopy(value)
        return RunePage.from_data(data)

    @transform.register(RunePagesData, RunePages)
    def rune_pages_data_to_core(self, value: RunePagesData, context: PipelineContext = None) -> RunePages:
        return RunePages([self.rune_page_data_to_core(page) for page in value], summoner=value.summoner_id)

    # Core to Dto

    @transform.register(RunePage, RunePageDto)
    def rune_page_core_to_dto(self, value: RunePage, context: PipelineContext = None) -> RunePageDto:
        return value._data[RunePageData]._dto

    @transform.register(RunePages, RunePagesDto)
    def rune_pages_core_to_dto(self, value: RunePages, context: PipelineContext = None) -> RunePagesDto:
        return RunePagesDto({"pages": set([self.rune_page_core_to_dto(page) for page in value]), "summonerId": value._RunePages__summoner.id})



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.spectator import CurrentGameInfoData, FeaturedGamesData, CurrentMatch, FeaturedMatches
from ..dto.spectator import CurrentGameInfoDto, FeaturedGamesDto

T = TypeVar("T")
F = TypeVar("F")


class SpectatorTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(CurrentGameInfoDto, CurrentGameInfoData)
    def current_game_dto_to_data(self, value: CurrentGameInfoDto, context: PipelineContext = None) -> CurrentGameInfoData:
        data = deepcopy(value)
        return CurrentGameInfoData.from_dto(data)

    @transform.register(FeaturedGamesDto, FeaturedGamesData)
    def featured_games_dto_to_data(self, value: FeaturedGamesDto, context: PipelineContext = None) -> FeaturedGamesData:
        data = deepcopy(value)
        data = data["gameList"]
        return FeaturedGamesData([CurrentGameInfoData.from_dto(game) for game in data], region=value["region"], client_refresh_interval=value["clientRefreshInterval"])

    # Data to Core

    @transform.register(CurrentGameInfoData, CurrentMatch)
    def current_game_data_to_core(self, value: CurrentGameInfoData, context: PipelineContext = None) -> CurrentMatch:
        from ..core.summoner import Summoner
        summoner = Summoner(name=value.teams[0].participants[0].summoner_name)
        return CurrentMatch.from_data(value, summoner=summoner)

    @transform.register(FeaturedGamesData, FeaturedMatches)
    def featured_games_data_to_core(self, value: FeaturedGamesData, context: PipelineContext = None) -> FeaturedMatches:
        from ..core.summoner import Summoner
        matches = []
        for match in value:
            summoner = Summoner(name=match.teams[0].participants[0].summoner_name)
            match = CurrentMatch.from_data(match, summoner=summoner)
            matches.append(match)
        return FeaturedMatches(matches, region=value.region, client_refresh_interval=value.client_refresh_interval)

    # Core to Dto

    @transform.register(CurrentMatch, CurrentGameInfoDto)
    def current_game_core_to_dto(self, value: CurrentMatch, context: PipelineContext = None) -> CurrentGameInfoDto:
        return value._data[CurrentGameInfoData]._dto

    @transform.register(FeaturedMatches, FeaturedMatches)
    def featured_games_core_to_dto(self, value: FeaturedGamesData, context: PipelineContext = None) -> FeaturedGamesDto:
        return FeaturedGamesDto({"clientRefreshInterval": value.client_refresh_interval, "gameList": [self.current_game_core_to_dto(game) for game in value], "region": value.region})



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.league import Leagues, LeaguesListData, LeagueListData, League, ChallengerLeagueListData, ChallengerLeague, MasterLeagueListData, MasterLeague

from ..dto.league import LeaguesListDto, LeagueListDto, ChallengerLeagueListDto, MasterLeagueListDto

T = TypeVar("T")
F = TypeVar("F")


class LeagueTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(LeagueListDto, LeagueListData)
    def league_dto_to_data(self, value: LeagueListDto, context: PipelineContext = None) -> LeagueListData:
        data = deepcopy(value)
        return LeagueListData.from_dto(data)

    @transform.register(LeaguesListDto, LeaguesListData)
    def leagues_dto_to_data(self, value: LeaguesListDto, context: PipelineContext = None) -> LeaguesListData:
        data = deepcopy(value)
        for league in data["leagues"]:
            league["region"] = data["region"]
            league["summonerId"] = data["summonerId"]
            for entry in league["entries"]:
                entry["tier"] = league["tier"]
                entry["leagueName"] = league["name"]
                entry["queue"] = league["queue"]
                entry["region"] = league["region"]
        data = [self.league_dto_to_data(league) for league in data["leagues"]]
        return LeaguesListData(data, summoner_id=value["summonerId"])

    @transform.register(ChallengerLeagueListDto, ChallengerLeagueListData)
    def challenger_league_dto_to_data(self, value: ChallengerLeagueListDto, context: PipelineContext = None) -> ChallengerLeagueListData:
        data = deepcopy(value)
        return ChallengerLeagueListData.from_dto(data)

    @transform.register(MasterLeagueListDto, MasterLeagueListData)
    def master_league_dto_to_data(self, value: MasterLeagueListDto, context: PipelineContext = None) -> MasterLeagueListData:
        data = deepcopy(value)
        return MasterLeagueListData.from_dto(data)

    # Data to Core

    @transform.register(LeagueListData, League)
    def league_data_to_core(self, value: LeagueListData, context: PipelineContext = None) -> League:
        data = deepcopy(value)
        return League.from_data(data)

    @transform.register(LeaguesListData, Leagues)
    def leagues_data_to_core(self, value: LeaguesListData, context: PipelineContext = None) -> Leagues:
        return Leagues([self.league_data_to_core(page) for page in value], summoner=value.summoner_id)

    @transform.register(ChallengerLeagueListData, ChallengerLeague)
    def challenger_league_data_to_core(self, value: ChallengerLeagueListData, context: PipelineContext = None) -> ChallengerLeague:
        data = deepcopy(value)
        return ChallengerLeague.from_data(data)

    @transform.register(MasterLeagueListData, MasterLeague)
    def master_league_data_to_core(self, value: MasterLeagueListData, context: PipelineContext = None) -> MasterLeague:
        data = deepcopy(value)
        return MasterLeague.from_data(data)

    # Core to Dto

    @transform.register(League, LeagueListDto)
    def league_core_to_dto(self, value: League, context: PipelineContext = None) -> LeaguesListDto:
        return value._data[LeagueListData]._dto

    @transform.register(Leagues, LeaguesListDto)
    def leagues_core_to_dto(self, value: Leagues, context: PipelineContext = None) -> LeaguesListDto:
        return LeaguesListDto({"leagues": set([self.league_core_to_dto(page) for page in value]), "summonerId": value._Leagues__summoner.id})

    @transform.register(ChallengerLeague, ChallengerLeagueListDto)
    def challenger_league_core_to_dto(self, value: ChallengerLeague, context: PipelineContext = None) -> ChallengerLeagueListDto:
        return value._data[ChallengerLeagueListData]._dto

    @transform.register(MasterLeague, MasterLeagueListDto)
    def master_league_core_to_dto(self, value: MasterLeague, context: PipelineContext = None) -> MasterLeagueListDto:
        return value._data[MasterLeagueListData]._dto



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.staticdata.champion import ChampionData, ChampionListData, Champion, Champions
from ..core.staticdata.mastery import MasteryData, MasteryListData, Mastery, Masteries
from ..core.staticdata.rune import RuneData, RuneListData, Rune, Runes
from ..core.staticdata.item import ItemData, ItemListData, Item, Items
from ..core.staticdata.summonerspell import SummonerSpellData, SummonerSpellListData, SummonerSpell, SummonerSpells
from ..core.staticdata.version import VersionListData, Versions
from ..core.staticdata.map import MapData, MapListData, Map, Maps
from ..core.staticdata.realm import RealmData, Realms
from ..core.staticdata.language import LanguagesData, Locales
from ..core.staticdata.languagestrings import LanguageStringsData, LanguageStrings
from ..core.staticdata.profileicon import ProfileIconData, ProfileIconListData, ProfileIcon, ProfileIcons

from ..dto.staticdata import ChampionDto, ChampionListDto
from ..dto.staticdata import MasteryDto, MasteryListDto
from ..dto.staticdata import RuneDto, RuneListDto
from ..dto.staticdata import ItemDto, ItemListDto
from ..dto.staticdata import SummonerSpellDto, SummonerSpellListDto
from ..dto.staticdata import VersionListDto
from ..dto.staticdata import MapDto, MapListDto
from ..dto.staticdata.realm import RealmDto
from ..dto.staticdata.language import LanguagesDto, LanguageStringsDto
from ..dto.staticdata.profileicon import ProfileIconDataDto, ProfileIconListDto

T = TypeVar("T")
F = TypeVar("F")


class StaticDataTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    ###############
    # Dto to Data #
    ###############

    # Champion

    @transform.register(ChampionDto, ChampionData)
    def champion_dto_to_data(self, value: ChampionDto, context: PipelineContext = None) -> ChampionData:
        data = deepcopy(value)
        return ChampionData.from_dto(data)

    @transform.register(ChampionListDto, ChampionListData)
    def champion_list_dto_to_data(self, value: ChampionListDto, context: PipelineContext = None) -> ChampionListData:
        data = deepcopy(value)

        data["data"] = [self.champion_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"], "includedData": data["includedData"]})

        data = data["data"]
        return ChampionListData(data, region=value["region"], version=value["version"], locale=value["locale"], included_data=value["includedData"])

    # Mastery

    @transform.register(MasteryDto, MasteryData)
    def mastery_dto_to_data(self, value: MasteryDto, context: PipelineContext = None) -> MasteryData:
        data = deepcopy(value)
        return MasteryData.from_dto(data)

    @transform.register(MasteryListDto, MasteryListData)
    def mastery_list_dto_to_data(self, value: MasteryListDto, context: PipelineContext = None) -> MasteryListData:
        data = deepcopy(value)

        data["data"] = [self.mastery_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"], "includedData": data["includedData"]})

        data = data["data"]
        return MasteryListData(data, region=value["region"], version=value["version"], locale=value["locale"], included_data=value["includedData"])

    # Rune

    @transform.register(RuneDto, RuneData)
    def rune_dto_to_data(self, value: RuneDto, context: PipelineContext = None) -> RuneData:
        data = deepcopy(value)
        return RuneData.from_dto(data)

    @transform.register(RuneListDto, RuneListData)
    def rune_list_dto_to_data(self, value: RuneListDto, context: PipelineContext = None) -> RuneListData:
        data = deepcopy(value)

        data["data"] = [self.rune_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"], "includedData": data["includedData"]})

        data = data["data"]
        return RuneListData(data, region=value["region"], version=value["version"], locale=value["locale"], included_data=value["includedData"])

    # Item

    @transform.register(ItemDto, ItemData)
    def item_dto_to_data(self, value: ItemDto, context: PipelineContext = None) -> ItemData:
        data = deepcopy(value)
        return ItemData.from_dto(data)

    @transform.register(ItemListDto, ItemListData)
    def item_list_dto_to_data(self, value: ItemListDto, context: PipelineContext = None) -> ItemListData:
        data = deepcopy(value)

        data["data"] = [self.item_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"], "includedData": data["includedData"]})

        data = data["data"]
        return ItemListData(data, region=value["region"], version=value["version"], locale=value["locale"], included_data=value["includedData"])

    # Summoner Spell

    @transform.register(SummonerSpellDto, SummonerSpellData)
    def summoner_spell_dto_to_data(self, value: SummonerSpellDto, context: PipelineContext = None) -> SummonerSpellData:
        data = deepcopy(value)
        return SummonerSpellData.from_dto(data)

    @transform.register(SummonerSpellListDto, SummonerSpellListData)
    def summoner_spell_list_dto_to_data(self, value: SummonerSpellListDto, context: PipelineContext = None) -> SummonerSpellListData:
        data = deepcopy(value)

        data["data"] = [self.summoner_spell_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"], "includedData": data["includedData"]})

        data = data["data"]
        return SummonerSpellListData(data, region=value["region"], version=value["version"], locale=value["locale"], included_data=value["includedData"])

    # Map

    @transform.register(MapDto, MapData)
    def map_dto_to_data(self, value: MapDto, context: PipelineContext = None) -> MapData:
        data = deepcopy(value)
        return MapData.from_dto(data)

    @transform.register(MapListDto, MapListData)
    def map_list_dto_to_data(self, value: MapListDto, context: PipelineContext = None) -> MapListData:
        data = deepcopy(value)

        data["data"] = [self.map_dto_to_data(c) for c in data["data"].values()]
        for c in data["data"]:
            c._update({"region": data["region"], "locale": data["locale"], "version": data["version"]})

        data = data["data"]
        return MapListData(data, region=value["region"], version=value["version"], locale=value["locale"])

    # Version

    @transform.register(VersionListDto, VersionListData)
    def version_list_dto_to_data(self, value: VersionListDto, context: PipelineContext = None) -> VersionListData:
        data = VersionListData(deepcopy(value["versions"]), region=value["region"])
        return data

    # Realm

    @transform.register(RealmDto, RealmData)
    def realm_dto_to_data(self, value: RealmDto, context: PipelineContext = None) -> RealmData:
        data = deepcopy(value)
        return RealmData(**data)

    # Languages

    @transform.register(LanguagesDto, LanguagesData)
    def languages_dto_to_data(self, value: LanguagesDto, context: PipelineContext = None) -> LanguagesData:
        data = deepcopy(value)
        return LanguagesData(data["languages"], region=value["region"])

    # Language Strings

    @transform.register(LanguageStringsDto, LanguageStringsData)
    def language_strings_dto_to_data(self, value: LanguageStringsDto, context: PipelineContext = None) -> LanguageStringsData:
        data = deepcopy(value)
        return LanguageStringsData(strings=data, region=value["region"], version=value["version"], locale=value["locale"])

    # Profile Icons

    @transform.register(ProfileIconDataDto, ProfileIconData)
    def profile_icon_dto_to_data(self, value: ProfileIconDataDto, context: PipelineContext = None) -> ProfileIconData:
        data = deepcopy(value)
        return ProfileIconData.from_dto(data)

    @transform.register(ProfileIconListDto, ProfileIconListData)
    def profile_icon_dto_to_data(self, value: ProfileIconListDto, context: PipelineContext = None) -> ProfileIconListData:
        data = deepcopy(value)
        return ProfileIconListData([self.profile_icon_dto_to_data(p) for p in data["data"]], region=value["region"], version=value["version"], locale=value["locale"])

    ################
    # Data to Core #
    ################

    # Champion

    @transform.register(ChampionData, Champion)
    def champion_data_to_core(self, value: ChampionData, context: PipelineContext = None) -> Champion:
        return Champion.from_data(value)

    @transform.register(ChampionListData, Champions)
    def champion_list_data_to_core(self, value: ChampionListData, context: PipelineContext = None) -> Champions:
        return Champions([self.champion_data_to_core(c) for c in value], region=value.region, version=value.version, locale=value.locale, included_data=value.included_data)

    # Mastery

    @transform.register(MasteryData, Mastery)
    def mastery_data_to_core(self, value: MasteryData, context: PipelineContext = None) -> Mastery:
        return Mastery.from_data(value)

    @transform.register(MasteryListData, Masteries)
    def mastery_list_data_to_core(self, value: MasteryListData, context: PipelineContext = None) -> Masteries:
        return Masteries([self.mastery_data_to_core(m) for m in value], region=value.region, version=value.version, locale=value.locale, included_data=value.included_data)

    # Rune

    @transform.register(RuneData, Rune)
    def rune_data_to_core(self, value: RuneData, context: PipelineContext = None) -> Rune:
        return Rune.from_data(value)

    @transform.register(RuneListData, Runes)
    def rune_list_data_to_core(self, value: RuneListData, context: PipelineContext = None) -> Runes:
        return Runes([self.rune_data_to_core(r) for r in value], region=value.region, version=value.version, locale=value.locale, included_data=value.included_data)

    # Item

    @transform.register(ItemData, Item)
    def item_data_to_core(self, value: ItemData, context: PipelineContext = None) -> Item:
        return Item.from_data(value)

    @transform.register(ItemListData, Items)
    def item_list_data_to_core(self, value: ItemListData, context: PipelineContext = None) -> Items:
        return Items([self.item_data_to_core(i) for i in value], region=value.region, version=value.version, locale=value.locale, included_data=value.included_data)

    # Summoner Spell

    @transform.register(SummonerSpellData, SummonerSpell)
    def summoner_spell_data_to_core(self, value: SummonerSpellData, context: PipelineContext = None) -> SummonerSpell:
        return SummonerSpell.from_data(value)

    @transform.register(SummonerSpellListData, SummonerSpells)
    def summoner_spell_list_data_to_core(self, value: SummonerSpellListData, context: PipelineContext = None) -> SummonerSpells:
        return SummonerSpells([self.summoner_spell_data_to_core(s) for s in value], region=value.region, version=value.version, locale=value.locale, included_data=value.included_data)

    # Map

    @transform.register(MapData, Map)
    def map_data_to_core(self, value: MapData, context: PipelineContext = None) -> Map:
        return Map.from_data(value)

    @transform.register(MapListData, Maps)
    def map_list_data_to_core(self, value: MapListData, context: PipelineContext = None) -> Maps:
        return Maps([self.map_data_to_core(m) for m in value], region=value.region, version=value.version, locale=value.locale)

    # Version

    @transform.register(VersionListData, Versions)
    def version_list_data_to_core(self, value: VersionListData, context: PipelineContext = None) -> Versions:
        version = Versions(value, region=value.region)
        return version

    # Realm

    @transform.register(RealmData, Realms)
    def realm_data_to_core(self, value: RealmData, context: PipelineContext = None) -> Realms:
        realms = Realms.from_data(value)
        realms(region=value.region)
        return realms

    # Languages

    @transform.register(LanguagesData, Locales)
    def languages_data_to_core(self, value: LanguagesData, context: PipelineContext = None) -> Locales:
        return Locales(value, region=value.region)

    # Language Strings

    @transform.register(LanguageStringsData, LanguageStrings)
    def language_strings_data_to_core(self, value: LanguageStringsData, context: PipelineContext = None) -> LanguageStrings:
        return LanguageStrings(strings=value.strings, region=value.region, version=value.version, locale=value.locale)

    # Profile Icon

    @transform.register(ProfileIconData, ProfileIcon)
    def profile_icon_data_to_core(self, value: ProfileIconData, context: PipelineContext = None) -> ProfileIcon:
        return ProfileIcon.from_data(value)

    @transform.register(ProfileIconListData, ProfileIcons)
    def profile_icon_data_to_core(self, value: ProfileIconListData, context: PipelineContext = None) -> ProfileIcons:
        return ProfileIcons(value, region=value.region, version=value.version, locale=value.locale)

    ###############
    # Core To Dto #
    ###############

    # Champion

    @transform.register(Champion, ChampionDto)
    def champion_core_to_dto(self, value: Champion, context: PipelineContext = None) -> ChampionDto:
        return value._data[ChampionData]._dto

    @transform.register(Champions, ChampionListDto)
    def champion_list_core_to_dto(self, value: Champions, context: PipelineContext = None) -> ChampionListDto:
        # I didn't put in `keys`, `type`, or `format` keys/values that come from the API.
        return ChampionListDto({"region": value.region, "version": value.version, "locale": value.locale, "includedData": value.included_data,
                                "data": {c.id: self.champion_core_to_dto(c) for c in value}})

    # Mastery

    @transform.register(Mastery, MasteryDto)
    def mastery_core_to_dto(self, value: Mastery, context: PipelineContext = None) -> MasteryDto:
        return value._data[MasteryData]._dto

    @transform.register(Masteries, MasteryListDto)
    def mastery_list_core_to_dto(self, value: Masteries, context: PipelineContext = None) -> MasteryListDto:
        return MasteryListDto({"region": value.region, "version": value.version, "locale": value.locale, "includedData": value.included_data,
                            "data": {m.id: self.mastery_core_to_dto(m) for m in value}})

    # Rune

    @transform.register(Rune, RuneDto)
    def rune_core_to_dto(self, value: Rune, context: PipelineContext = None) -> RuneDto:
        return value._data[RuneData]._dto

    @transform.register(Runes, RuneListDto)
    def rune_list_core_to_dto(self, value: Runes, context: PipelineContext = None) -> RuneListDto:
        return RuneListDto({"region": value.region, "version": value.version, "locale": value.locale, "includedData": value.included_data,
                            "data": {r.id: self.rune_core_to_dto(r) for r in value}})

    # Item

    @transform.register(Item, ItemDto)
    def item_core_to_dto(self, value: Item, context: PipelineContext = None) -> ItemDto:
        return value._data[ItemData]._dto

    @transform.register(Items, ItemListDto)
    def item_list_core_to_dto(self, value: Items, context: PipelineContext = None) -> ItemListDto:
        return ItemListDto({"region": value.region, "version": value.version, "locale": value.locale, "includedData": value.included_data,
                            "data": {i.id: self.item_core_to_dto(i) for i in value}})

    # Summoner Spell

    @transform.register(SummonerSpell, SummonerSpellDto)
    def summoner_spell_core_to_dto(self, value: SummonerSpell, context: PipelineContext = None) -> SummonerSpellDto:
        return value._data[SummonerSpellData]._dto

    @transform.register(SummonerSpells, SummonerSpellListDto)
    def summoner_spell_list_core_to_dto(self, value: SummonerSpells, context: PipelineContext = None) -> SummonerSpellListDto:
        return SummonerSpellListDto({"region": value.region, "version": value.version, "locale": value.locale, "includedData": value.included_data,
                            "data": {s.key: self.summoner_spell_core_to_dto(s) for s in value}})

    # Map

    @transform.register(Map, MapDto)
    def map_core_to_dto(self, value: Map, context: PipelineContext = None) -> MapDto:
        return value._data[MapData]._dto

    @transform.register(Maps, MapListDto)
    def map_list_core_to_dto(self, value: Maps, context: PipelineContext = None) -> MapListDto:
        return MapListDto({"region": value.region, "version": value.version, "locale": value.locale, "data": {str(m.id): self.map_core_to_dto(m) for m in value}})

    # Version

    @transform.register(Versions, VersionListDto)
    def version_list_core_to_dto(self, value: Versions, context: PipelineContext = None) -> VersionListDto:
        return VersionListDto({"region": value.region, "data": [v for v in value]})

    # Realm

    @transform.register(Realms, RealmDto)
    def realm_core_to_dto(self, value: Realms, context: PipelineContext = None) -> RealmDto:
        return RealmDto({"region": value.region, **value[RealmData]._dto})

    # Languages

    @transform.register(Locales, LanguagesDto)
    def languages_core_to_dto(self, value: Locales, context: PipelineContext = None) -> LanguagesDto:
        return LanguagesDto({"region": value.region, "languages": [v for v in value]})

    # Language Strings

    @transform.register(LanguageStrings, LanguageStringsDto)
    def language_strings_core_to_dto(self, value: LanguageStrings, context: PipelineContext = None) -> LanguageStringsDto:
        # I left out `type`.
        return LanguageStringsDto({"region": value.region, "version": value.version, "locale": value.locale, **value[LanguageStringsData]._dto})

    # Profile Icon

    @transform.register(ProfileIcon, ProfileIconDataDto)
    def profile_icon_core_to_dto(self, value: ProfileIcon, context: PipelineContext = None) -> ProfileIconDataDto:
        return value._data[ProfileIconData]._dto

    @transform.register(ProfileIcons, ProfileIconListDto)
    def profile_icon_core_to_dto(self, value: ProfileIcons, context: PipelineContext = None) -> ProfileIconListDto:
        # I left out `type`.
        return ProfileIconListDto({"region": value.region, "version": value.version, "locale": value.locale, "data": [self.profile_icon_core_to_dto(m) for m in value]})
